/*
 * drivers/mtd/nand/ts7400.c
 *
 * Copyright (C) 2004 Technologic Systems (support@embeddedARM.com)
 *
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/sizes.h>
#include <asm/mach-types.h>


struct partition_info { // little-endian format
  unsigned int bootable:8; // 0x80 = active, 0x00 = don't use for booting
  unsigned int starting_head:8;
  unsigned int starting_csect:8;
  unsigned int starting_xcyl:8;
  // MSB bits 0-5 are sector
  unsigned int system_id:8; // format of partition
  unsigned int ending_head:8;
  unsigned int ending_csect:8;
  unsigned int ending_xcyl:8;
  unsigned int relative_sector:32; // sector offset from start of disk to start of volume
  unsigned int total_sectors:32;
} __attribute__((packed));

struct partition_table {
  struct partition_info part[4];
} __attribute__((__packed__));

struct MBR {
  char bootcode[446];
  struct partition_table pt;
  unsigned int sig55:8;
  unsigned int sigAA:8;
} __attribute__((__packed__));

/*
 * MTD structure for TS7250 board
 */
static struct mtd_info *ts7400_mtd = NULL;

#define MBRBUFF 512
#define NUM_PARTITIONS 3

static volatile unsigned long addr_control;
static volatile unsigned long addr_dat;
static volatile unsigned long addr_busy;


/*
static struct nand_ecclayout nand_oob_16 = {
        .eccbytes = 6,
        .eccpos = {0, 1, 2, 3, 6, 7},
        .oobfree = { {8, 8} }
};
*/
/*
static struct nand_oobinfo nand_oob_64 = {
	.useecc = MTD_NANDECC_AUTOPLACE,
	.eccbytes = 24,
	.eccpos = {
		40, 41, 42, 43, 44, 45, 46, 47, 
		48, 49, 50, 51, 52, 53, 54, 55, 
		56, 57, 58, 59, 60, 61, 62, 63},
	.oobfree = { {2, 38} }
};
*/

static struct mtd_partition partition_mbr[] = {
	{ .name= "whole chip",
		  .offset= 0,
		  .size= 0x2000000 },
	{ .name= "kernel",
		  .offset= 0x4000,
		  .size= 0x200000 },
	{ .name= "initrd",
		  .offset= 0x204000,
		  .size= 0x200000 },
	{ .name= "rootfs",
		  .offset= 0x404000,
		  .size= 0x2000000 - 0x404000 },
	{ .name= "part4",
		  .offset= 0x2000000,
		  .size= 0 },
};


/*
 *	hardware specific access to control-lines
 *
 *	ctrl:
 *	NAND_NCE: bit 0 -> bit 2
 *	NAND_CLE: bit 1 -> bit 1
 *	NAND_ALE: bit 2 -> bit 0
 */
static void ts7400_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	struct nand_chip *chip = mtd->priv;

	if (ctrl & NAND_CTRL_CHANGE) {
		unsigned long addr =  addr_control;
		unsigned char bits;

		bits = (ctrl & NAND_NCE) << 2;
		bits |= ctrl & NAND_CLE;
		bits |= (ctrl & NAND_ALE) >> 2;

		__raw_writeb((__raw_readb(addr) & ~0x7) | bits, addr);
	}

	if (cmd != NAND_CMD_NONE)
		__raw_writeb(cmd, chip->IO_ADDR_W);
}

/*
static int detect_hwecc(struct nand_chip *this) {

  struct ts7XX0 *myts;
  int dummy;
  volatile int *ts7x;

  myts = this->priv;	
  ts7x = (int *)myts->addr_control;
  dummy = *ts7x;
  dummy = *ts7x;
  
  //printk("Dummy 0x%x\n",dummy);
  if ((dummy >> 8) != -1) {
    return 0;
  }
  writeb(0x1, this->IO_ADDR_W);
  //dummy =  readw(myts->addr_control);
  dummy = *ts7x;
  return ((dummy >> 8) != -1);
}

static int ts7XX0_calculate_ecc(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code) {

     struct nand_chip *chip = mtd->priv;
     struct ts7XX0 *myts;
     int dummy;	
     volatile int *ts7x;
     
     myts = chip->priv;
     ts7x = (int *) myts->addr_control;
     dummy =  0xFFFFFF & ((*ts7x >> 6) | 3);
     
     //printk("Hardware ECC: %06X ",dummy);
     ecc_code[0] = (dummy >> 16) & 0xFF;
     ecc_code[1] = (dummy >> 8) & 0xFF;
     ecc_code[2] = (dummy >> 0) & 0xFF;
     //printk("(%02X %02X %02X)\n",ecc_code[0],ecc_code[1],ecc_code[2]);
     return dummy;
}

static void ts7400_enable_hwecc(struct mtd_info *mtd,int dir) {

  struct nand_chip *chip = mtd->priv;
  struct ts7XX0 *myts;
  int dummy;
  volatile int *ts7x;

  myts = chip->priv;	
  ts7x = (int *)myts->addr_control;
  dummy = *ts7x;

 // printk("enable_hwecc\n");
}
*/


/*
 *	read device ready pin
 */
static int ts7400_device_ready(struct mtd_info *mtd)
{
	return __raw_readb(addr_control)&0x20;
}

/*
 * Main initialization routine
 */
static int __init ts7400_init(void)
{
	struct nand_chip *this;
	char *nandctrl, *nanddat;
	const char *part_type = 0;
	char buf[MBRBUFF];
	int i,got= 0,parts=0,hwecc =0;
	unsigned ofs=0;
	struct MBR *mbr = (struct MBR *)buf;

	if (!machine_is_ts72xx() || board_is_ts7200())
		return -ENXIO;

	/* Allocate memory for MTD device structure and private data */
	ts7400_mtd = kmalloc(sizeof(struct mtd_info) + 
                             sizeof(struct nand_chip), GFP_KERNEL);
	if (!ts7400_mtd) {
		printk("Unable to allocate TS7400 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *)(&ts7400_mtd[1]);

	/* Initialize structures */
	memset(ts7400_mtd, 0, sizeof(struct mtd_info));
	memset(this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	ts7400_mtd->priv = this;
	ts7400_mtd->owner = THIS_MODULE;

	nandctrl = (char *)ioremap(0x60400000,0x100);
	nanddat  = (char *)ioremap(0x60000000,0x100);

	/* insert callbacks */
        addr_control = (unsigned long)nandctrl;
	addr_dat = (unsigned long)nanddat;
	this->IO_ADDR_R = nanddat + 0x2;
	this->IO_ADDR_W = nanddat + 0x0;
	this->cmd_ctrl = ts7400_hwcontrol;
	this->dev_ready = ts7400_device_ready;
	this->chip_delay = 15;
	this->options = NAND_USE_FLASH_BBT;
	this->ecc.mode = NAND_ECC_SOFT;
	
	/*
        if((hwecc = detect_hwecc(this)))
	{
		this->ecc.mode = NAND_ECC_HW;
		this->ecc.size = 256;
		this->ecc.bytes = 3;
		//this->ecc.layout = &nand_oob_16;
		this->ecc.hwctl = ts7250_enable_hwecc;
		this->ecc.calculate = ts7XX0_calculate_ecc;
		this->ecc.correct = nand_correct_data;
	}
	*/
	
	printk("Searching for NAND flash...\n");
	/* Scan to find existence of the device */
	if (nand_scan(ts7400_mtd, 1)) {
		kfree(ts7400_mtd);
		return -ENXIO;
	}
	
	printk("Hardware ECC");
	if(!hwecc) printk(" NOT");	
	printk(" detected!\n");

	//null mbrbuffer
	for (i=0;i<MBRBUFF;i++) {
	  buf[i] = 0;
	}

	ts7400_mtd->read(ts7400_mtd,0,MBRBUFF,&got,buf);
	if (mbr->sigAA == 0xAA && mbr->sig55 == 0x55) {
		part_type = "MBR";
		for (i=0;i<4;i++) {
			if (mbr->pt.part[i].total_sectors) {
				parts++;
				partition_mbr[i+1].offset = 
				512 * mbr->pt.part[i].relative_sector;
				if (i < 2) {
					partition_mbr[i+1].size = 
					512 * mbr->pt.part[i+1].relative_sector -
					512 * mbr->pt.part[i].relative_sector;
				} else {
					partition_mbr[i+1].size = 
					512 * mbr->pt.part[i].total_sectors;
				}
				ofs = partition_mbr[i+1].offset + partition_mbr[i+1].size;
			}
		}
	} else {
		part_type = "default";
		parts = 3;
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(ts7400_mtd, partition_mbr, parts+1);

	/* Return happy */
	return 0;
}

module_init(ts7400_init);

/*
 * Clean up routine
 */
static void __exit ts7400_cleanup(void)
{
	/* Unregister the device */
	del_mtd_device(ts7400_mtd);
	iounmap((void *)addr_control);
	iounmap((void *)addr_dat);

	/* Free the MTD device structure */
	kfree(ts7400_mtd);
}

module_exit(ts7400_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jesse Off <joff@embeddedARM.com>");
MODULE_DESCRIPTION("MTD map driver for Technologic Systems TS-7250 board");

