/*
 *  linux/drivers/serial/tsuart7370.c
 *
 *  Linux 2.6 Driver for Technologic Systems UART
 *
 *  (c) 2007 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2008-04-14:mos:initial version, based on tsuart7350.c
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
#include <asm/uaccess.h>
#include "tsuart1.h"

//---------------------------------------------------------------------------
// TS-7370 magic 0x6C92
//---------------------------------------------------------------------------
#define BOARD_ID_7370 0x7370
#define FPGA_BASE 0x600ff080
volatile unsigned short *tmp;

static int mdmctrl(struct tsuart_port *ts,int mctrl) {
  int dtr,dcd,port;

  // DTR is an output
  // DCD is an input
  dtr = (mctrl & TIOCM_DTR) ? 1 : 0;
  dcd = (mctrl & TIOCM_CAR) ? 1 : 0;

  port = (((int)ts->u.mapbase) - 0x600ff0a0)/4;
  //                  DATA OE   DTR    DCD
  // DTR=B8,  DCD=B11 0x0C 0x14 0x0080 0x0400
  // DTR=B15, DCD=B15 0x0C 0x14 0x8000 0x4000
  if (port == 9) {
    if (mctrl & 0x80000000) {
      if (dtr) {
	tmp[0x0C/2] |= 0x0080;
      } else {
	tmp[0x0C/2] &= ~0x0080;
      }
    }
    dtr = !!(tmp[0x0C/2] & 0x0080);
    dcd = !!(tmp[0x0C/2] & 0x0400);
  } else if (port == 10) {
    if (mctrl & 0x80000000) {
      if (dtr) {
	tmp[0x0C/2] |= 0x8000;
      } else {
	tmp[0x0C/2] &= ~0x8000;
      }
    }
    dtr = !!(tmp[0x0C/2] & 0x8000);
    dcd = !!(tmp[0x0C/2] & 0x4000);
  }
  return ((dcd ? TIOCM_CAR : 0) | TIOCM_DSR | (dtr ? TIOCM_DTR : 0));
}

struct uart_driver *dr9=0;

static int tryInitPort(int offset) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  memset(ts,0,sizeof(struct tsuart_port));

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetMEM16;
  ts->put = tsuartPutMEM16;
  ts->mdmctrl = (offset > 8) ? mdmctrl : 0;
  ts->debugFlags = portDebug;
  ts->u.mapbase = 0x600ff0a0 + 4 * offset;
  ts->u.irq = 32;
//  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_7370;

  if (tsuart_register_port(0,ts) != 0) {
    printk("Failed to register port!\n");
    return 0;
  }

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart nine-bit port\n");
    return 0;
  }
  memset(ts,0,sizeof(struct tsuart_port));

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetMEM16_9;
  ts->put = tsuartPutMEM16_9;
  ts->mdmctrl = (offset > 8) ? mdmctrl : 0;
  ts->debugFlags = portDebug;
  ts->pstatus |= PORT_STATUS_TXMSB;
  ts->rxsize = 2;
  ts->u.mapbase = 0x600ff0a0 + 4 * offset;
  ts->u.irq = 32;
//  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_7370;

  if (tsuart_register_port(dr9,ts) != 0) {
    printk("Failed to register 9-bit port!\n");
    return 0;
  }

  return 1;
}

//---------------------------------------------------------------------------
// KERNEL MODULE
//---------------------------------------------------------------------------
static int __init tsuart_init(void)
{
  int i;

  dr9 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr9) {
    printk("Unable to allocate memory for tsuart port\n");
    return -1;
  }

  tsuart_configure_driver(dr9,234,65,"tt8s",12);
  tryInitPort(3);
  tryInitPort(4);
  tryInitPort(0);

  return 0;
}

static void __exit tsuart_exit(void)
{
}

module_init(tsuart_init);
module_exit(tsuart_exit);

MODULE_DESCRIPTION("Technologic Systems TS-UART driver");


